#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h> 
#include <linux/videodev2.h>
#include <sys/mman.h>
#include <string.h>

#define DEV "/dev/video2"
//#define WIDTH 640
//#define HEIGHT 480
#define BUF_NUM 4
#define BRIGHTNESS 10	// 设置亮度
#define CONTRAST 10		// 设置对比度

struct pic_data
{
	unsigned char *tmpbuffer[BUF_NUM];
	unsigned int tmpbytesused[BUF_NUM];
}pic;

int fd;
int type;
int jpg_fd;
int jpg_size;
char file_name[50];
char cap_format[50];
char *fmt_suffix = "yuv";
int cap_pixelfmt = V4L2_PIX_FMT_YUYV;
int cap_num = 1;
char cap_path[100] = "photo";
float sleep_time = 1000*1000;
int WIDTH = 640;
int HEIGHT = 480;

int v4l2_init()
{
	// 1. 打开设备
	if ((fd = open(DEV, O_RDWR | O_NONBLOCK)) < 0)
	{
		perror("ERROR open device");
		return -1;
	}

	
	// 2. 查询设备能力
	struct v4l2_capability cap;
	if (ioctl(fd, VIDIOC_QUERYCAP, &cap) < 0)
	{
		perror("ERROR querycap");
		return -1;
	}
	if ((cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0)
	{
		perror("ERROR the device is not supported capture");
		return -1;
	}
	
	// 枚举摄像头支持格式
	struct v4l2_fmtdesc fmtdesc;
	fmtdesc.index = 0;
	fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    printf("Support format:\n");
    while(ioctl(fd, VIDIOC_ENUM_FMT, &fmtdesc) != -1) {
        printf("\t%d.%s\n", fmtdesc.index+1, fmtdesc.description);
        fmtdesc.index++;
    }
    
	// 枚举分辨率
	struct v4l2_frmsizeenum frmsize;
	frmsize.index = 0;
	frmsize.pixel_format = cap_pixelfmt;
	frmsize.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	printf("支持的分辨率：\n");
	while (ioctl(fd, VIDIOC_ENUM_FRAMESIZES, &frmsize) != -1)
	{
		printf("%d. %d*%d\n", frmsize.index, frmsize.discrete.width, frmsize.discrete.height);
		frmsize.index++;
	}
	
	// 3. 设置格式
	struct v4l2_format fmt;
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width = 480;
	fmt.fmt.pix.height = 320;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
	fmt.fmt.pix.field = V4L2_FIELD_ANY;
	int ret = ioctl(fd, VIDIOC_S_FMT, &fmt);
	if (ret < 0)
	{
		printf ("ret = %d\n", ret);
		perror("ERROR set format");
		return -1;
	}
	// 4. 查询格式设置是否成功
	if (ioctl(fd, VIDIOC_G_FMT, &fmt) < 0)
	{
		perror("ERROR get format");
		return -1;
	}
	if (fmt.fmt.pix.pixelformat == cap_pixelfmt)
	{
		printf("set format succeed!\n");
	}
	/*
	// 5. 设置控制参数
	struct v4l2_control ctrl;
	ctrl.id = V4L2_CID_BRIGHTNESS;	// 设置亮度
	ctrl.value = BRIGHTNESS;
	ctrl.id = V4L2_CID_CONTRAST;	// 设置对比度
	ctrl.value = CONTRAST;
	if (ioctl(fd, VIDIOC_S_CTRL, &ctrl) < 0)
	{
		perror("ERROR set ctrl parameters");
		return -1;
	}
	*/
	// 6. 申请缓冲
	struct v4l2_requestbuffers reqbuf;
	reqbuf.count = BUF_NUM;
	reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	if (ioctl(fd, VIDIOC_REQBUFS, &reqbuf) < 0)
	{
		perror("ERROR request buffer");
		return -1;
	}
	
	// 7. 内存映射
	struct v4l2_buffer buf;
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;
	for (int i = 0; i < BUF_NUM; i++)
	{
		buf.index = i;
		if (ioctl(fd, VIDIOC_QUERYBUF, &buf) < 0)
		{
			perror("ERROR query buffer");
			return -1;
		}
		
		pic.tmpbuffer[i] = (unsigned char*)mmap(NULL, buf.length, PROT_READ|PROT_WRITE, MAP_SHARED, fd, buf.m.offset);
	}
	
	// 8. 入队
	for (int i = 0; i < BUF_NUM; i++)
	{
		buf.index = i;
		if (ioctl(fd, VIDIOC_QBUF, &buf) < 0)
		{
			perror("ERROR queue buffer");
			return -1;
		}
	}
	
	// 9. 开始传输数据
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(fd, VIDIOC_STREAMON, &type) < 0)
	{
		perror("ERROR stream on");
		return -1;
	}
	
	return 0;
}

int v4l2_capture(int i)
{
	printf("start of DQBUF!!!\n");
	// 10. 读取图像数据，出队
	struct v4l2_buffer buf;
	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;
	if (ioctl(fd, VIDIOC_DQBUF, &buf) < 0)
	{
		perror("ERROR dequeue buffer");
		return -1;
	}
	printf("end of DQBUF!!!\n");
	
	pic.tmpbytesused[buf.index] = buf.bytesused;
	//printf("Size : %d\n", pic.tmpbytesused[buf.index]);
	// 11. 保存图片
//	snprintf(file_name, 200,"%s/cap%d.%s", cap_path, i, fmt_suffix);
	snprintf(file_name, 200,"cap%d.%s", i, fmt_suffix);
	jpg_fd = open(file_name, O_RDWR|O_CREAT, 0777);
	if (jpg_fd < 0)
	{
		perror("ERROR open cap.jpeg");
		return -1;
	}
	i++;
	
	jpg_size = write(jpg_fd, pic.tmpbuffer[buf.index], pic.tmpbytesused[buf.index]);
	printf("%d. Write Successfully size : %d\n", i, jpg_size);
	close(jpg_fd);
	// 12. 重新入队
	if (ioctl(fd, VIDIOC_QBUF, &buf) < 0)
	{
		perror("ERROR again queue buffer");
		return -1;
	}

	return 0;
}

int v4l2_close()
{
	// 13. 停止传输数据
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(fd, VIDIOC_STREAMOFF, &type) < 0)
	{
		perror("ERROR stream off");
		return -1;
	}
	// 14. 取消映射
	for (int i = 0; i < BUF_NUM; i++)
	{
		munmap(pic.tmpbuffer[i], pic.tmpbytesused[i]);
	}
	close(fd);
	
	return 0;
}

int main(int argc, char *argv[])
{/*
	if (argc == 2 && strcmp(argv[1], "-s") == 0)
	{
		//printf("请输入图片的保存路径（默认保存在photo目录下）：\n");
		//scanf("%s", cap_path);
		
		printf("请输入保存的图片张数：\n");
		scanf("%d", &cap_num);
		
		printf("请输入拍照的时间间隔（单位：秒）：\n");
		scanf("%f", &sleep_time);
		sleep_time *= 1000*1000; 
		
		printf("请输入拍照格式（MJPEG 或 YUYV）:\n");
		scanf("%s", cap_format);
		if (strcmp(cap_format ,"MJPEG") == 0)
		{
			cap_pixelfmt = V4L2_PIX_FMT_MJPEG;
			fmt_suffix = "jpeg";
		}
		else if (strcmp(cap_format ,"YUYV") == 0)
		{
			cap_pixelfmt = V4L2_PIX_FMT_YUYV;
			fmt_suffix = "yuv";
		}
		else
		{
			printf("请输入正确的格式！！！\n");
			return -1;
		}
	}
	*/
	v4l2_init();
	for (int i = 0; i < cap_num; i++)
	{
		//printf("i = %d, cap_num = %d\n", i, cap_num);
		v4l2_capture(i);
		usleep(sleep_time);
	}
	v4l2_close();
	return 0;
}
